@@ -5,10 +5,13 @@ import concurrent.futures
 import copy
 import logging
 import os
+from typing import Any, Dict, cast
 
 from ..constants import MimeType
 from ..decoding import decode_data as decode_data_function
 from ..io_utils import read_data, write_data
+from ..type_utils import JsonDict
+from .request import DownloadRequest
 from .sentinelhub_client import SentinelHubDownloadClient
 
 LOGGER = logging.getLogger(__name__)
@@ -22,21 +25,19 @@ class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
 
     _RETRIABLE_ERRORS = ["EXECUTION_ERROR", "TIMEOUT"]
 
-    def __init__(self, *args, n_interval_retries=1, max_retry_threads=5, **kwargs):
+    def __init__(self, *args: Any, n_interval_retries: int = 1, max_retry_threads: int = 5, **kwargs: Any):
         """
         :param n_interval_retries: Number of retries if a request fails just for a certain timestamp. (This parameter
             is experimental and might be changed in the future.)
-        :type n_interval_retries: int
         :param max_retry_threads: Number of threads used for retrying. (This parameter is experimental and might be
             changed in the future.)
-        :type max_retry_threads: int
         """
         super().__init__(*args, **kwargs)
 
         self.n_interval_retries = n_interval_retries
         self.max_retry_threads = max_retry_threads
 
-    def _single_download(self, request, _):
+    def _single_download(self, request: DownloadRequest, _: Any) -> Any:
         """Method for downloading a single request"""
         request.raise_if_invalid()
         if not (request.save_response or request.return_data):
@@ -49,11 +50,12 @@ class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
             response_content = self._execute_download(request)
             stats_response = decode_data_function(response_content, request.data_type)
         else:
-            LOGGER.debug("Reading locally stored data from %s instead of downloading", response_path)
+            path = cast(str, response_path)  # fix in future, for now ensured by download_required
+            LOGGER.debug("Reading locally stored data from %s instead of downloading", path)
             self._check_cached_request_is_matching(request, request_path)
-            stats_response = read_data(response_path, data_format=request.data_type)
+            stats_response = read_data(path, data_format=request.data_type)
 
-        failed_time_intervals = {}
+        failed_time_intervals: Dict[int, Any] = {}
         for index, stat_info in enumerate(stats_response["data"]):
             if self._has_retriable_error(stat_info):
                 failed_time_intervals[index] = stat_info["interval"]
@@ -70,22 +72,26 @@ class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
 
         if request_path and request.save_response and (self.redownload or not os.path.exists(request_path)):
             request_info = request.get_request_params(include_metadata=True)
-            write_data(request_path, request_info, data_format=MimeType.JSON)
+            write_data(request_path, request_info, MimeType.JSON)
             LOGGER.debug("Saved request info to %s", request_path)
 
         if request.save_response and (download_required or n_succeeded_intervals > 0):
-            write_data(response_path, stats_response, data_format=request.data_type)
+            path = cast(str, response_path)  # fix in future, for now ensured by raise_if_invalid
+            write_data(path, stats_response, data_format=request.data_type)
             LOGGER.debug("Saved data to %s", response_path)
 
         if request.return_data:
             return stats_response
         return None
 
-    def _download_per_interval(self, request, time_intervals):
+    def _download_per_interval(self, request: DownloadRequest, time_intervals: Dict[int, Any]) -> dict:
         """Download statistics per each time interval"""
         interval_requests = []
         for time_interval in time_intervals.values():
             interval_request = copy.deepcopy(request)
+            if interval_request.post_values is None or "aggregation" not in interval_request.post_values:
+                raise ValueError("Unable to configure request for retrying by interval.")
+
             interval_request.post_values["aggregation"]["timeRange"] = time_interval
             interval_requests.append(interval_request)
 
@@ -94,7 +100,7 @@ class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
 
         return dict(zip(time_intervals, stat_info_responses))
 
-    def _execute_single_stat_download(self, request):
+    def _execute_single_stat_download(self, request: DownloadRequest) -> Any:
         """Makes sure a download for a single time interval is retried"""
         for retry_count in range(self.n_interval_retries):
             response = self._execute_download(request)
@@ -106,7 +112,7 @@ class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
 
         raise ValueError("No more retries available, download unsuccessful")
 
-    def _has_retriable_error(self, stat_info):
+    def _has_retriable_error(self, stat_info: JsonDict) -> bool:
         """Checks if a dictionary of Stat API info for a single time interval has an error that can fixed by retrying
         a request
         """
